def input_qasm(circ_str: str):
    """Process an input qasm string
    A simple qasm parser that works on many qasm files that I know of.
    The current OLSQ formulation only support single and two-qubitgates,
    so only those are processed.  This parser has some more assumptions:
    1) each instruction in a newline, 2) two-qubit gates: cx or zz,
    and 3) there is only one array of qubits.

    Args:
        circ_str: a string in qasm format.

    Returns:
        a list [qubit_num, gates, gate_spec]
            qubit_num: logical qubit number.
            gates: which qubit(s) each gate acts on.
            gate_spec: type of each gate.

    Example:
        for the following circuit 
            q_0: ───────────────────■───
                                    │  
            q_1: ───────■───────────┼───
                 ┌───┐┌─┴─┐┌─────┐┌─┴─┐
            q_2: ┤ H ├┤ X ├┤ TDG ├┤ X ├─
                 └───┘└───┘└─────┘└───┘ 
        gates = ((2,), (1,2), (2,), (0,1))
        gate_spec = ("h", "cx", "tdg", "cx")
    """

    gates = list() # which qubit(s) each gate acts on
    gate_spec = list() # type of each gate
    qubit_num = 0

    for qasmline in circ_str.splitlines():
        words = qasmline.split()
        if not isinstance(words, list): continue # empty lines
        if len(words) == 0: continue
        grammar = len(words)
        # grammer=2 -> single-qubit_gate_name one_qubit
        #              |two-qubit_gate_name one_qubit,one_qubit
        # grammer=3 ->  two-qubit_gate_name one_qubit, one_qubit
        # grammer=1 or >3 incorrect, raise an error
        
        if words[0] in ["OPENQASM", "include", "creg", "measure", "//"]:
            continue

        if words[0] == 'qreg':
            qubit_num = int(words[1][2:-2])
            continue
        
        if words[0] == 'cx' or words[0] == 'zz':
            if grammar == 3:
                try:
                    qubit0 = int(words[1][2:-2])
                    qubit1 = int(words[2][2:-2])
                except:
                    raise ValueError(f"{qasmline} invalid two-qubit gate.")
            
            elif grammar == 2:
                qubits = words[1].split(',')
                try:
                    qubit0 = int(qubits[0][2:-1])
                    qubit1 = int(qubits[1][2:-2])
                except:
                    raise ValueError(f"{qasmline} invalid two-qubit gate.")
            
            else:
                raise ValueError(f"{qasmline}: invalid two-qubit gate.")
            
            gates.append((qubit0, qubit1))
            gate_spec.append(words[0])
        
        elif grammar == 2: # not two-qubit gate, and has two words
            try:
                qubit = int(words[1][2:-2])
            except:
                raise ValueError(f"{qasmline} invalid single-qubit gate.")
            gates.append((qubit,))
            gate_spec.append(words[0])
        
        else:
            raise ValueError(f"{qasmline} invalid gate.")

    if qubit_num == 0:
        raise ValueError("Qubit number is not specified.")
    return [qubit_num, tuple(gates), tuple(gate_spec)]



from cirq import Circuit


def input_cirq(circuit: Circuit, map_physical_qubit_to: dict):
    """Handle a Cirq circuit object as input.
    In Cirq, the qubits are objects, in OLSQ IR, qubits are integers, so
    we need to store a mapping from indices to Cirq qubits.

    Args:
        circuit: a Cirq Circuit object
        map_physical_qubit_to: a mapping from Cirq qubits to indices

    Returns:
        count_program_qubit: in OLSQ IR
        list_gate_qubits: in OLSQ IR
        list_gate_name: in OLSQ IR, not merely string, but Cirq objects
    """

    list_gate_qubits = []
    list_gate_name = []

    for operation in circuit.all_operations():
        list_gate_name.append(operation.gate)
        if len(operation.qubits) == 1:
            list_gate_qubits.append(
                (map_physical_qubit_to[operation.qubits[0]],) )
        elif len(operation.qubits) == 2:
            list_gate_qubits.append(
                (map_physical_qubit_to[operation.qubits[0]],
                 map_physical_qubit_to[operation.qubits[1]]  )  )
        else:
            raise ValueError("Only support single-qubit and two-qubit gates.")

    return [len(circuit.all_qubits()), tuple(list_gate_qubits),
            tuple(list_gate_name)]



from cirq import Circuit, Moment, GateOperation, SWAP, CNOT
import networkx as nx

from olsq.solve import OLSQ
from olsq.device import qcdevice
from olsq.olsq_cirq.input import input_cirq


class OLSQ_cirq(OLSQ):
    def __init__(self, objective_name, if_transition_based):
        """Set the objective of OLSQ_cirq, and whether transition-based.

        Args:
            objective_name: can be "depth", "swap", or "fidelity"
            mode: can be "normal" or "transition" (TB-OLSQ in the paper)       
        """

        super().__init__(objective_name, if_transition_based)
        self.map_physical_qubit_to = dict() # Cirq qubit |-> qubit index
        self.map_to_physical_qubit = []     # qubit index |-> Cirq qubit

    def setdevicegraph(self, device_graph: nx.Graph):
        """Pass in a graph representing hardware in Cirq, also generate
        a mapping from indices to physical qubits, and reverse mapping.

        Args:
            device_graph: an NetworkX Graph object where vertices are
                Cirq qubits and edges mean physical connections.
                Some code in Cirq use this representation of QC devices.
        """

        count_physical_qubit = len(list(device_graph.nodes()))
        for i, qubit in enumerate(list(device_graph.nodes())):
            self.map_to_physical_qubit.append(qubit)
            self.map_physical_qubit_to[qubit] = i

        qubit_edges = []
        for edge in list(device_graph.edges()):
            qubit_edges.append( (self.map_physical_qubit_to[edge[0]],
                                 self.map_physical_qubit_to[edge[1]]) )
        if self.if_transition_based:
            tmp = 1
        else:
            tmp = 3
        super().setdevice(
            qcdevice("cirq-dev", nqubits=count_physical_qubit,
                     connection=tuple(qubit_edges), swap_duration=tmp) )

    def setprogram(self, program_cirq: Circuit):
        """Translate input Cirq program/circuit to IR

        Args:
            program_cirq: a Circuit object in Cirq
        """

        program_cirq = input_cirq(program_cirq, self.map_physical_qubit_to)
        super().setprogram(program_cirq, input_mode="IR")

    def solve(self):
        """Use the super().solve(...) and output result in Cirq

        Returns:
            circuit: result in Cirq Circuit object
            final__cirq_mapping: from input Cirq qubit to Cirq qubit
            objective_value: depth/#swap/fidelity depending on setting

        Note:
            final_mapping (returned by super().solve(...)): 
                logical qubit index |-> physical qubit index
            map_to_physical_qubit: qubit index |-> Cirq qubit
            So the final mapping should be m(i) |-> m(f(i))
        """

        (result_depth, list_scheduled_gate_name, list_scheduled_gate_qubits,
            final_mapping, objective_value) = super().solve(output_mode="IR")
        
        # constructing the output Cirq Circuit object
        circuit = Circuit()
        for t in range(result_depth):
            moment_with_swap = []
            for i, qubits in enumerate(list_scheduled_gate_qubits[t]):
                if len(qubits) == 1:
                    moment_with_swap.append( 
                        GateOperation( list_scheduled_gate_name[t][i],
                            (self.map_to_physical_qubit[qubits[0]],)   ) )
                else:
                    if list_scheduled_gate_name[t][i] == "SWAP":
                        moment_with_swap.append(
                            GateOperation(SWAP,
                                (self.map_to_physical_qubit[qubits[0]],
                                 self.map_to_physical_qubit[qubits[1]]) ) )
                    else:
                        op_tmp = list_scheduled_gate_name[t][i]
                        if list_scheduled_gate_name[t][i] == "cx":
                            op_tmp = CNOT
                        moment_with_swap.append(
                            GateOperation(op_tmp,
                                (self.map_to_physical_qubit[qubits[0]],
                                 self.map_to_physical_qubit[qubits[1]])   ) )
            circuit.append(moment_with_swap)

        final_cirq_mapping = dict()
        for i in range(self.count_program_qubit):
            final_cirq_mapping[ self.map_to_physical_qubit[i] ] = \
                self.map_to_physical_qubit[ final_mapping[i] ]

        return [circuit, final_cirq_mapping, objective_value]



import math
import datetime

from z3 import Int, IntVector, Bool, Optimize, Implies, And, Or, If, sat

from olsq.input import input_qasm
from olsq.output import output_qasm
from olsq.device import qcdevice
import pkgutil


def collision_extracting(list_gate_qubits):
    """Extract collision relations between the gates,
    If two gates g_1 and g_2 both acts on a qubit (at different time),
    we say that g_1 and g_2 collide on that qubit, which means that
    (1,2) will be in collision list.

    Args:
        list_gate_qubits: a list of gates in OLSQ IR
    
    Returns:
        list_collision: a list of collisions between the gates
    """

    list_collision = list()
    # We sweep through all the gates.  For each gate, we sweep through all the
    # gates after it, if they both act on some qubit, append them in the list.
    for g in range(len(list_gate_qubits)):
        for gg in range(g + 1, len(list_gate_qubits)):
            
            if list_gate_qubits[g][0] == list_gate_qubits[gg][0]:
                    list_collision.append((g, gg))
                
            if len(list_gate_qubits[gg]) == 2:
                if list_gate_qubits[g][0] == list_gate_qubits[gg][1]:
                    list_collision.append((g, gg))
            
            if len(list_gate_qubits[g]) == 2:
                if list_gate_qubits[g][1] == list_gate_qubits[gg][0]:
                    list_collision.append((g, gg))
                if len(list_gate_qubits[gg]) == 2:
                    if list_gate_qubits[g][1] == list_gate_qubits[gg][1]:
                        list_collision.append((g, gg))
    
    return tuple(list_collision)

def dependency_extracting(list_gate_qubits, count_program_qubit: int):
    """Extract dependency relations between the gates.
    If two gates g_1 and g_2 both acts on a qubit *and there is no gate
    between g_1 and g_2 that act on this qubit*, we then say that
    g2 depends on g1, which means that (1,2) will be in dependency list.

    Args:
        list_gate_qubits: a list of gates in OLSQ IR
        count_program_qubit: the number of logical/program qubit
    
    Returns:
        list_dependency: a list of dependency between the gates
    """

    list_dependency = []
    list_last_gate = [-1 for i in range(count_program_qubit)]
    # list_last_gate records the latest gate that acts on each qubit.
    # When we sweep through all the gates, this list is updated and the
    # dependencies induced by the update is noted.
    for i, qubits in enumerate(list_gate_qubits):
        
        if list_last_gate[qubits[0]] >= 0:
            list_dependency.append((list_last_gate[qubits[0]], i))
        list_last_gate[qubits[0]] = i

        if len(qubits) == 2:
            if list_last_gate[qubits[1]] >= 0:
                list_dependency.append((list_last_gate[qubits[1]], i))
            list_last_gate[qubits[1]] = i

    return tuple(list_dependency)


class OLSQ:
    def __init__(self, objective_name, mode):
        """Set the objective of OLSQ, and whether it is transition-based

        Args:
            objective_name: can be "depth", "swap", or "fidelity"
            mode: can be "normal" or "transition" (TB-OLSQ in the paper)       
        """
        
        if objective_name == "depth":
            self.objective_name = objective_name
        elif objective_name == "swap":
            self.objective_name = objective_name
        elif objective_name == "fidelity":
            self.objective_name = objective_name
        else:
            raise ValueError("Invalid Objective Name")

        if mode == "transition":
            self.if_transition_based = True
        elif mode == "normal":
            self.if_transition_based = False
        else:
            raise ValueError("Invalid Choice of Transition-Based or Not")

        # These values should be updated in setdevice(...)
        self.device = None
        self.count_physical_qubit = 0
        self.list_qubit_edge = []
        self.swap_duration = 0

        # These values should be updated in setprogram(...)
        self.list_gate_qubits = []
        self.count_program_qubit = 0
        self.list_gate_name = []
        
        # bound_depth is a hyperparameter
        self.bound_depth = 0

        self.inpput_dependency = False
        self.list_gate_dependency = []

    def setdevice(self, device: qcdevice):
        """Pass in parameters from the given device.  If in TB mode,
           swap_duration is set to 1 without modifying the device.

        Args:
            device: a qcdevice object for OLSQ
        """

        self.device = device
        self.count_physical_qubit = device.count_physical_qubit
        self.list_qubit_edge = device.list_qubit_edge
        self.swap_duration = device.swap_duration
        if self.if_transition_based:
            self.swap_duration = 1

    def setprogram(self, program, input_mode: str = None):
        """Translate input program to OLSQ IR, and set initial depth
        An example of the intermediate representation is shown below.
        It contains three things: 1) the number of qubit in the program,
        2) a list of tuples representing qubit(s) acted on by a gate,
        the tuple has one index if it is a single-qubit gate,
        two indices if it is a two-qubit gate, and 3) a list of
        type/name of each gate, which is not important to OLSQ,
        and only needed when generating output.
        If in TB mode, initial depth=1; in normal mode, we perform ASAP
        scheduling without consideration of SWAP to calculate depth.

        Args:
            program: a qasm string, or a list of the three things in IR.
            input_mode: (optional) can be "IR" if the input has ben
                translated to OLSQ IR; can be "benchmark" to use one of
                the benchmarks.  Default mode assumes qasm input.

        Example:
            For the following circuit
                q_0: ───────────────────■───
                                        │  
                q_1: ───────■───────────┼───
                     ┌───┐┌─┴─┐┌─────┐┌─┴─┐
                q_2: ┤ H ├┤ X ├┤ TDG ├┤ X ├─
                     └───┘└───┘└─────┘└───┘ 
            count_program_qubit = 3
            gates = ((2,), (1,2), (2,), (0,1))
            gate_spec = ("h", "cx", "tdg", "cx")
        """
        
        if input_mode == "IR":
            self.count_program_qubit = program[0]
            self.list_gate_qubits = program[1]
            self.list_gate_name = program[2]
        elif input_mode == "benchmark":
            f = pkgutil.get_data(__name__, "benchmarks/" + program + ".qasm")
            program = input_qasm(f.decode("utf-8"))
            self.count_program_qubit = program[0]
            self.list_gate_qubits = program[1]
            self.list_gate_name = program[2]
        else:
            program = input_qasm(program)
            self.count_program_qubit = program[0]
            self.list_gate_qubits = program[1]
            self.list_gate_name = program[2]

        # calculate the initial depth
        if self.if_transition_based:
            self.bound_depth = 1
        else:
            push_forward_depth = [0 for i in range(self.count_program_qubit)]
            for qubits in self.list_gate_qubits:
                if len(qubits) == 1:
                    push_forward_depth[qubits[0]] += 1
                else:
                    tmp_depth = push_forward_depth[qubits[0]]
                    if tmp_depth < push_forward_depth[qubits[1]]:
                        tmp_depth = push_forward_depth[qubits[1]]
                    push_forward_depth[qubits[1]] = tmp_depth + 1
                    push_forward_depth[qubits[0]] = tmp_depth + 1
            self.bound_depth = max(push_forward_depth)

    def setdependency(self, dependency: list):
        """Specify dependency (non-commutation)

        Args:
            dependency: a list of gate index pairs
        
        Example:
            For the following circuit
                q_0: ───────────────────■───
                                        │  
                q_1: ───────■───────────┼───
                     ┌───┐┌─┴─┐┌─────┐┌─┴─┐
                q_2: ┤ H ├┤ X ├┤ TDG ├┤ X ├─
                     └───┘└───┘└─────┘└───┘ 
                gate   0    1     2     3
            dependency = [(0,1), (1,2), (2,3)]

            However, for this QAOA subcircuit (ZZ gates may have phase
            parameters, but we neglect them for simplicity here)
                         ┌──┐ ┌──┐
                q_0: ────┤ZZ├─┤  ├─
                     ┌──┐└┬─┘ │ZZ│  
                q_1: ┤  ├─┼───┤  ├─
                     │ZZ│┌┴─┐ └──┘
                q_2: ┤  ├┤ZZ├──────
                     └──┘└──┘ 
                gate   0   1   2
            dependency = []    # since ZZ gates are commutable
        """
        self.list_gate_dependency = dependency
        self.inpput_dependency = True

    def solve(self, output_mode: str = None, output_file_name: str = None):
        """Formulate an SMT, pass it to z3 solver, and output results.
        CORE OF OLSQ, EDIT WITH CARE.

        Args:
            output_mode: "IR" or left to default.
            output_file_name: a file to store the IR output or qasm.
        
        Returns:
            a list of results depending on output_mode
            "IR": 
            | list_scheduled_gate_name: name/type of each gate
            | list_scheduled_gate_qubits: qubit(s) each gate acts on
            | final_mapping: logical qubit |-> physical qubit in the end 
            | objective_value: depth/#swap/fidelity depending on setting
            None:
              a qasm string
              final_mapping
              objective_value
        """

        objective_name = self.objective_name
        device = self.device
        list_gate_qubits = self.list_gate_qubits
        count_program_qubit = self.count_program_qubit
        list_gate_name = self.list_gate_name
        count_physical_qubit = self.count_physical_qubit
        list_qubit_edge = self.list_qubit_edge
        swap_duration = self.swap_duration
        bound_depth = self.bound_depth

        # pre-processing

        count_qubit_edge = len(list_qubit_edge)
        count_gate = len(list_gate_qubits)
        if self.objective_name == "fidelity":
            list_logfidelity_single = [
                int(1000 * math.log(device.list_fidelity_single[n]))
                for n in range(count_physical_qubit)]
            list_logfidelity_two = [
                int(1000 * math.log(device.list_fidelity_two[k]))
                for k in range(count_qubit_edge)]
            list_logfidelity_measure = [
                int(1000 * math.log(device.list_fidelity_measure[n]))
                for n in range(count_physical_qubit)]
        list_gate_two = list()
        list_gate_single = list()
        for l in range(count_gate):
            if len(list_gate_qubits[l]) == 1:
                list_gate_single.append(l)
            else:
                list_gate_two.append(l)

        # list_adjacency_qubit takes in a physical qubit index _p_, and
        # returns the list of indices of physical qubits adjacent to _p_
        list_adjacent_qubit = list()
        # list_span_edge takes in a physical qubit index _p_,
        # and returns the list of edges spanned from _p_
        list_span_edge = list()
        for n in range(count_physical_qubit):
            list_adjacent_qubit.append(list())
            list_span_edge.append(list())
        for k in range(count_qubit_edge):
            list_adjacent_qubit[list_qubit_edge[k][0]].append(
                                                        list_qubit_edge[k][1])
            list_adjacent_qubit[list_qubit_edge[k][1]].append(
                                                        list_qubit_edge[k][0])
            list_span_edge[list_qubit_edge[k][0]].append(k)
            list_span_edge[list_qubit_edge[k][1]].append(k)

        # if_overlap_edge takes in two edge indices _e_ and _e'_,
        # and returns whether or not they overlap
        if_overlap_edge = [[0] * count_qubit_edge
            for k in range(count_qubit_edge)]
        # list_over_lap_edge takes in an edge index _e_,
        # and returnsthe list of edges that overlap with _e_
        list_overlap_edge = list()
        # list_count_overlap_edge is the list of lengths of
        # overlap edge lists of all the _e_
        list_count_overlap_edge = list()
        for k in range(count_qubit_edge):
            list_overlap_edge.append(list())
        for k in range(count_qubit_edge):
            for kk in range(k + 1, count_qubit_edge):
                if (   (list_qubit_edge[k][0] == list_qubit_edge[kk][0]
                        or list_qubit_edge[k][0] == list_qubit_edge[kk][1])
                    or (list_qubit_edge[k][1] == list_qubit_edge[kk][0]
                        or list_qubit_edge[k][1] == list_qubit_edge[kk][1]) ):
                    list_overlap_edge[k].append(kk)
                    list_overlap_edge[kk].append(k)
                    if_overlap_edge[kk][k] = 1
                    if_overlap_edge[k][kk] = 1
        for k in range(count_qubit_edge):
            list_count_overlap_edge.append(len(list_overlap_edge[k]))

        if not self.inpput_dependency:
            list_gate_dependency = collision_extracting(list_gate_qubits)
        else:
            list_gate_dependency = self.list_gate_dependency

        # index function: takes two physical qubit indices _p_ and _p'_,
        # and returns the index of the edge between them if there is one
        map_edge_index = [[0] * count_physical_qubit] * count_physical_qubit
        for k in range(count_qubit_edge):
            map_edge_index[list_qubit_edge[k][0]][list_qubit_edge[k][1]] = k
            map_edge_index[list_qubit_edge[k][1]][list_qubit_edge[k][0]] = k

        not_solved = True
        start_time = datetime.datetime.now()
        while not_solved:
            print("Trying maximal depth = {}...".format(bound_depth))

            # variable setting 

            # at cycle t, logical qubit q is mapped to pi[q][t]
            pi = [[Int("map_q{}_t{}".format(i, j)) for j in range(bound_depth)]
                  for i in range(count_program_qubit)]

            # time coordinate for gate l is time[l]
            time = IntVector('time', count_gate)

            # space coordinate for gate l is space[l]
            space = IntVector('space', count_gate)

            # if at cycle t, a SWAP finishing on edge k, then sigma[k][t]=1
            sigma = [[Bool("ifswap_e{}_t{}".format(i, j))
                for j in range(bound_depth)] for i in range(count_qubit_edge)]

            # for depth optimization
            depth = Int('depth')

            # for swap optimization
            count_swap = Int('num_swap')

            # for fidelity optimization
            if objective_name == "fidelity":
                u = [Int("num_1qbg_p{}".format(n))
                     for n in range(count_physical_qubit)]
                v = [Int("num_2qbg_e{}".format(k))
                     for k in range(count_qubit_edge)]
                vv = [Int("num_swap_e{}".format(k))
                      for k in range(count_qubit_edge)]
                w = [Int("num_meas_p{}".format(n))
                     for n in range(count_physical_qubit)]
                fidelity = Int('log_fidelity')

            lsqc = Optimize()

            # constraint setting

            for t in range(bound_depth):
                for m in range(count_program_qubit):
                    lsqc.add(pi[m][t] >= 0, pi[m][t] < count_physical_qubit)
                    for mm in range(m):
                        lsqc.add(pi[m][t] != pi[mm][t])

            for l in range(count_gate):
                lsqc.add(time[l] >= 0, time[l] < bound_depth)
                if l in list_gate_single:
                    lsqc.add(space[l] >= 0, space[l] < count_physical_qubit)
                    for t in range(bound_depth):
                        lsqc.add(Implies(time[l] == t,
                            pi[list_gate_qubits[l][0]][t] == space[l]))
                elif l in list_gate_two:
                    lsqc.add(space[l] >= 0, space[l] < count_qubit_edge)
                    for k in range(count_qubit_edge):
                        for t in range(bound_depth):
                            lsqc.add(Implies(And(time[l] == t, space[l] == k),
                                Or(And(list_qubit_edge[k][0] == \
                                        pi[list_gate_qubits[l][0]][t],
                                    list_qubit_edge[k][1] == \
                                        pi[list_gate_qubits[l][1]][t]),
                                And(list_qubit_edge[k][1] == \
                                        pi[list_gate_qubits[l][0]][t],
                                    list_qubit_edge[k][0] == \
                                        pi[list_gate_qubits[l][1]][t])  )    ))

            for d in list_gate_dependency:
                if self.if_transition_based:
                    lsqc.add(time[d[0]] <= time[d[1]])
                else:
                    lsqc.add(time[d[0]] < time[d[1]])

            for t in range(min(swap_duration - 1, bound_depth)):
                for k in range(count_qubit_edge):
                    lsqc.add(sigma[k][t] == False)

            for t in range(swap_duration - 1, bound_depth):
                for k in range(count_qubit_edge):
                    for tt in range(t - swap_duration + 1, t):
                        lsqc.add(Implies(sigma[k][t] == True,
                            sigma[k][tt] == False))
                    for tt in range(t - swap_duration + 1, t + 1):
                        for kk in list_overlap_edge[k]:
                            lsqc.add(Implies(sigma[k][t] == True,
                                sigma[kk][tt] == False))

            if not self.if_transition_based:
                for t in range(swap_duration - 1, bound_depth):
                    for k in range(count_qubit_edge):
                        for tt in range(t - swap_duration + 1, t + 1):
                            for l in range(count_gate):
                                if l in list_gate_single:
                                    lsqc.add(Implies(And(time[l] == tt,
                                        Or(space[l] == list_qubit_edge[k][0],
                                           space[l] == list_qubit_edge[k][1])),
                                            sigma[k][t] == False             ))
                                elif l in list_gate_two:
                                    lsqc.add(Implies(And(
                                        time[l] == tt, space[l] == k),
                                            sigma[k][t] == False           ))
                                    for kk in list_overlap_edge[k]:
                                        lsqc.add(Implies(And(
                                            time[l] == tt, space[l] == kk),
                                                sigma[k][t] == False       ))

            for t in range(bound_depth - 1):
                for n in range(count_physical_qubit):
                    for m in range(count_program_qubit):
                        lsqc.add(
                            Implies(And(sum([If(sigma[k][t], 1, 0)
                                for k in list_span_edge[n]]) == 0,
                                    pi[m][t] == n), pi[m][t + 1] == n))

            for t in range(bound_depth - 1):
                for k in range(count_qubit_edge):
                    for m in range(count_program_qubit):
                        lsqc.add(Implies(And(sigma[k][t] == True,
                            pi[m][t] == list_qubit_edge[k][0]),
                                pi[m][t + 1] == list_qubit_edge[k][1]))
                        lsqc.add(Implies(And(sigma[k][t] == True,
                            pi[m][t] == list_qubit_edge[k][1]),
                                pi[m][t + 1] == list_qubit_edge[k][0]))

            lsqc.add(
                count_swap == sum([If(sigma[k][t], 1, 0)
                    for k in range(count_qubit_edge)
                        for t in range(bound_depth)]))

            # for depth optimization
            for l in range(count_gate):
                lsqc.add(depth >= time[l] + 1)

            if objective_name == "swap":
                lsqc.minimize(count_swap)
            elif objective_name == "depth":
                lsqc.minimize(depth)
            elif objective_name == "fidelity":
                for n in range(count_physical_qubit):
                    lsqc.add(u[n] == sum([If(space[l] == n, 1, 0)
                        for l in list_gate_single]))
                    lsqc.add(w[n] == sum([If(pi[m][bound_depth - 1] == n, 1, 0)
                        for m in range(count_program_qubit)]))
                for k in range(count_qubit_edge):
                    lsqc.add(v[k] == sum([If(space[l] == k, 1, 0)
                        for l in list_gate_two]))
                    lsqc.add(vv[k] == sum([If(sigma[k][t], 1, 0)
                        for t in range(bound_depth)]))

                lsqc.add(fidelity == sum([v[k] * list_logfidelity_two[k]
                    for k in range(count_qubit_edge)])
                    + sum([
                        swap_duration * vv[k] * list_logfidelity_two[k]
                        for k in range(count_qubit_edge)])
                    + sum([w[n] * list_logfidelity_measure[n]
                        for n in range(count_physical_qubit)])
                    + sum([u[n] * list_logfidelity_single[n]
                        for n in range(count_physical_qubit)])          )
                lsqc.maximize(fidelity)
            else:
                raise Exception("Invalid Objective Name")

            satisfiable = lsqc.check()
            if satisfiable == sat:
                not_solved = False
            else:
                if self.if_transition_based:
                    bound_depth += 1
                else:
                    bound_depth = int(1.3 * bound_depth)

        print(f"Compilation time = {datetime.datetime.now() - start_time}.")
        model = lsqc.model()

        # post-processing

        result_time = []
        result_depth = model[depth].as_long()
        for l in range(count_gate):
            result_time.append(model[time[l]].as_long())
        list_result_swap = []
        for k in range(count_qubit_edge):
            for t in range(result_depth):
                if model[sigma[k][t]]:
                    list_result_swap.append((k, t))
                    print(f"SWAP on physical edge ({list_qubit_edge[k][0]},"\
                        + f"{list_qubit_edge[k][1]}) at time {t}")
        for l in range(count_gate):
            if len(list_gate_qubits[l]) == 1:
                qq = list_gate_qubits[l][0]
                tt = result_time[l]
                print(f"Gate {l}: {list_gate_name[l]} {qq} on qubit "\
                    + f"{model[pi[qq][tt]].as_long()} at time {tt}")
            else:
                qq = list_gate_qubits[l][0]
                qqq = list_gate_qubits[l][1]
                tt = result_time[l]
                print(f"Gate {l}: {list_gate_name[l]} {qq}, {qqq} on qubits "\
                    + f"{model[pi[qq][tt]].as_long()} and "\
                    + f"{model[pi[qqq][tt]].as_long()} at time {tt}")


        # transition based
        if self.if_transition_based:
            self.swap_duration = self.device.swap_duration
            real_time = [0 for i in range(count_gate)]
            list_depth_on_qubit = [-1 for i in range(self.count_physical_qubit)]
            list_real_swap = []
            for block in range(result_depth):
                for tmp_gate in range(count_gate):
                    if result_time[tmp_gate] == block:
                        qubits = list_gate_qubits[tmp_gate]
                        if len(qubits) == 1:
                            p0 = model[pi[qubits[0]][block]].as_long()
                            real_time[tmp_gate] = \
                                list_depth_on_qubit[p0] + 1
                            list_depth_on_qubit[p0] = \
                                real_time[tmp_gate]
                        else:
                            p0 = model[pi[qubits[0]][block]].as_long()
                            p1 = model[pi[qubits[1]][block]].as_long()
                            real_time[tmp_gate] = max(
                                list_depth_on_qubit[p0],
                                list_depth_on_qubit[p1]) + 1
                            list_depth_on_qubit[p0] = \
                                real_time[tmp_gate]
                            list_depth_on_qubit[p1] = \
                                real_time[tmp_gate]
                            # print(f"{tmp_gate} {p0} {p1} real-time={real_time[tmp_gate]}")

                if block < result_depth - 1:
                    for (k, t) in list_result_swap:
                        if t == block:
                            p0 = list_qubit_edge[k][0]
                            p1 = list_qubit_edge[k][1]
                            tmp_time = max(list_depth_on_qubit[p0],
                                list_depth_on_qubit[p1]) \
                                + self.swap_duration
                            list_depth_on_qubit[p0] = tmp_time
                            list_depth_on_qubit[p1] = tmp_time
                            list_real_swap.append((k, tmp_time))
                # print(list_depth_on_qubit)
            result_time = real_time
            real_depth = 0
            for tmp_depth in list_depth_on_qubit:
                if real_depth < tmp_depth + 1:
                    real_depth = tmp_depth + 1
            result_depth = real_depth
            list_result_swap = list_real_swap
            # print(list_result_swap)
        
        objective_value = 0
        if objective_name == "fidelity":
            objective_value = model[fidelity].as_long()
            print(f"result fidelity = {math.exp(objective_value / 1000.0)}")
        elif objective_name == "swap":
            objective_value = len(list_result_swap)
            print(f"result additional SWAP count = {objective_value}.")
        else:
            objective_value = model[depth].as_long()
            print(f"result circuit depth = {objective_value}.")

        list_scheduled_gate_qubits = [[] for i in range(result_depth)]
        list_scheduled_gate_name = [[] for i in range(result_depth)]
        result_depth = 0
        for l in range(count_gate):
            t = result_time[l]

            if result_depth < t + 1:
                result_depth = t + 1

            list_scheduled_gate_name[t].append(list_gate_name[l])
            if l in list_gate_single:
                q = model[space[l]].as_long()
                list_scheduled_gate_qubits[t].append((q,))
            elif l in list_gate_two:
                [q0, q1] = list_gate_qubits[l]
                tmp_t = t
                if self.if_transition_based:
                    tmp_t = model[time[l]].as_long()
                q0 = model[pi[q0][tmp_t]].as_long()
                q1 = model[pi[q1][tmp_t]].as_long()
                list_scheduled_gate_qubits[t].append((q0, q1))
            else:
                raise ValueError("Expect single-qubit or two-qubit gate.")

        final_mapping = []
        for m in range(count_program_qubit):
            tmp_depth = result_depth - 1
            if self.if_transition_based:
                tmp_depth = model[depth].as_long() - 1
            final_mapping.append(model[pi[m][tmp_depth]].as_long())

        for (k, t) in list_result_swap:
            q0 = list_qubit_edge[k][0]
            q1 = list_qubit_edge[k][1]
            if self.swap_duration == 1:
                list_scheduled_gate_qubits[t].append((q0, q1))
                list_scheduled_gate_name[t].append("SWAP")
            elif self.swap_duration == 3:
                list_scheduled_gate_qubits[t].append((q0, q1))
                list_scheduled_gate_name[t].append("cx")
                list_scheduled_gate_qubits[t - 1].append((q1, q0))
                list_scheduled_gate_name[t - 1].append("cx")
                list_scheduled_gate_qubits[t - 2].append((q0, q1))
                list_scheduled_gate_name[t - 2].append("cx")
            else:
                raise ValueError("Expect SWAP duration one, or three")

        if output_mode == "IR":
            if output_file_name:
                output_file = open(output_file_name, 'w')
                output_file.writelines([list_scheduled_gate_name,
                                        list_scheduled_gate_qubits,
                                        final_mapping,
                                        objective_value])
            return (result_depth,
                    list_scheduled_gate_name,
                    list_scheduled_gate_qubits,
                    final_mapping,
                    objective_value)
        else:
            return (output_qasm(device, result_depth, list_scheduled_gate_name,
                                list_scheduled_gate_qubits, final_mapping,
                                True, output_file_name),
                    final_mapping,
                    objective_value)



from olsq.device import qcdevice


def output_qasm(device: qcdevice, result_depth: int,
                list_scheduled_gate_name: list,
                list_scheduled_gate_qubits: list,
                final_mapping: list, if_measure: bool, file_name: str = None):
    """Generate qasm output
    By default: 1) the qasm sentences are separated by newline,
    2) there is a comment before every moment in the circuit, and
    3) from the measurement, we can read of the final mapping. 
    If in the original circuit (input to OLSQ) a qubit has index i,
    then it will be measured at c[i] in the qasm string.

    Args:
        device: use the same QC device as in OLSQ.setdevice()
        result_depth: depth of the scheduled circuit
        list_scheduled_gate_name: type/name of each gate
        list_scheduled_gate_qubits: which qubit(s) each gate acts on
        final_mapping: mapping from logical qubits to physical qubits
        if_measure: if measure all qubits in the end
        file_name: (optional) a file to put the qasm string

    Returns:
        output_str: a qasm string in the above mentioned format.
    """

    count_physical_qubit = device.count_physical_qubit
    swap_duration = device.swap_duration
    
    # a list of strings, each string represents a moment in the circuit
    list_moment = ["" for i in range(result_depth)]
    for t in range(result_depth):
        for j, qubits in enumerate(list_scheduled_gate_qubits[t]):
            if len(qubits) == 1:
                list_moment[t] += list_scheduled_gate_name[t][j]
                list_moment[t] += f" q[{qubits[0]}];\n"
            else:
                list_moment[t] += list_scheduled_gate_name[t][j]
                list_moment[t] += f" q[{qubits[0]}], q[{qubits[1]}];\n"

    output_str = "OPENQASM 2.0;\ninclude \"qelib1.inc\";\n"
    output_str += f"qreg q[{count_physical_qubit}];\n"
    output_str += f"creg c[{count_physical_qubit}];\n"
    for t in range(result_depth):
        output_str += f"\n// moment {t}\n"
        output_str += list_moment[t]
    
    if if_measure:
        output_str += "\n// measurement\n"
        for program_qubit, physical_qubit in enumerate(final_mapping):
            output_str += f"measure q[{physical_qubit}]->c[{program_qubit}];\n"

    if file_name is not None:
        output_file = open(file_name, "w")
        output_file.write(output_str)

    return output_str



import json
import pkgutil


class qcdevice:
    """ QC device class.
    Contains the necessary parameters of the quantum hardware for OLSQ.
    """

    def __init__(self, name: str, nqubits: int = None, connection: list = None,
                 swap_duration: int = None, fmeas: list = None, 
                 fsingle: list = None, ftwo: list = None):
        """ Create a QC device.
        The user can either input the device parameters, or use existing
        ones stored in olsq/devices/ in json format (especially for
        duplicating paper results).  The parameters of existing devices 
        are overriden if inputs are provided.

        Args:
            name: name for the device.  If it starts with "default_",
                use existing device; otherwise, more parameters needed.
            nqubits: (optional) the number of physical qubits.
            connection: (optional) set of edges connecting qubits.
            swap_duration: (optional) how many time units a SWAP takes.
            fmeas: (optional) measurement fidelity of each qubit.
            fsingle: (optional) single-qubit gate fidelity of each qubit
            ftwo: (optional) two-qubit gate fidelity of each edge.

        Example:
            To use existing "defualt_ourense" device
            >>> dev = qcdevice(name="default_ourense")
            To set up a new device
            >>> dev = qcdevice(name="dev", nqubits=5,
                    connection=[(0, 1), (1, 2), (1, 3), (3, 4)],
                    swap_duration=3)
        """

        # typechecking for inputs
        if not isinstance(name, str):
            raise TypeError("name should be a string.")
        if nqubits is not None:
            if not isinstance(nqubits, int):
                raise TypeError("nqubits should be an integer.")
        if swap_duration is not None:
            if not isinstance(swap_duration, int):
                raise TypeError("swap_duration should be an integer.")
        
        if connection is not None:
            if not isinstance(connection, (list, tuple)):
                raise TypeError("connection should be a list or tuple.")
            else:
                for edge in connection:
                    if not isinstance(edge, (list, tuple)):
                        raise TypeError(f"{edge} is not a list or tuple.")
                    elif len(edge) != 2:
                        raise TypeError(f"{edge} does not connect two qubits.")
                    if not isinstance(edge[0], int):
                        raise TypeError(f"{edge[0]} is not an integer.")
                    if not isinstance(edge[1], int):
                        raise TypeError(f"{edge[1]} is not an integer.")
        
        if fmeas is not None:
            if not isinstance(fmeas, (list, tuple)):
                raise TypeError("fmeas should be a list or tuple.")
            else:
                for fmeas_i in fmeas:
                    if not isinstance(fmeas_i, (int, float)):
                        raise TypeError(f"{fmeas_i} is not a number.")
        if fsingle is not None:
            if not isinstance(fsingle, (list, tuple)):
                raise TypeError("fsingle should be a list or tuple.")
            else:
                for fsingle_i in fsingle:
                    if not isinstance(fsingle_i, (int, float)):
                        raise TypeError(f"{fsingle_i} is not a number.")
        if ftwo is not None:
            if not isinstance(ftwo, (list, tuple)):
                raise TypeError("ftwo should be a list or tuple.")
            else:
                for ftwo_i in ftwo:
                    if not isinstance(ftwo_i, (int, float)):
                        raise TypeError(f"{ftwo_i} is not a number.")
        
        if name.startswith("default_"):
            # use an existing device
            f = pkgutil.get_data(__name__, "devices/" + name + ".json")
            data = json.loads(f)
            self.name = data["name"]
            self.count_physical_qubit = data["count_physical_qubit"]
            self.list_qubit_edge = tuple( tuple(edge)
                                          for edge in data["list_qubit_edge"])
            self.swap_duration = data["swap_duration"]
            if "list_fidelity_measure" in data:
                self.list_fidelity_measure = \
                    tuple(data["list_fidelity_measure"])
            if "list_fidelity_single" in data:
                self.list_fidelity_single = tuple(data["list_fidelity_single"])
            if "list_fidelity_two" in data:
                self.list_fidelity_two = tuple(data["list_fidelity_two"])
        else:
            self.name = name
        
        # set parameters from inputs with value checking
        if nqubits is not None:
            self.count_physical_qubit = nqubits
        if "count_physical_qubit" not in self.__dict__:
            raise AttributeError("No physical qubit count specified.")

        if connection is not None:
            for edge in connection:
                if edge[0] < 0 or edge[0] >= self.count_physical_qubit:
                    raise ValueError( (f"{edge[0]} is outside of range "
                                       f"[0, {self.count_physical_qubit}).") )
                if edge[1] < 0 or edge[1] >= self.count_physical_qubit:
                    raise ValueError( (f"{edge[1]} is outside of range "
                                       f"[0, {self.count_physical_qubit}).") )
            self.list_qubit_edge = tuple( tuple(edge) for edge in connection)
        if "list_qubit_edge" not in self.__dict__:
            raise AttributeError("No edge set is specified.")
        
        if swap_duration is not None: 
            self.swap_duration = swap_duration
        else:
            self.swap_duration = 3
        
        if fmeas is not None:
            if len(fmeas) != self.count_physical_qubit:
                raise ValueError( ("fmeas should have "
                                   f"{self.count_physical_qubit} data.") )
            self.list_fidelity_measure = tuple(fmeas)
        if "list_fidelity_measure" not in self.__dict__:
            self.list_fidelity_measure = \
                tuple(1 for _ in range(self.count_physical_qubit))
        
        if fsingle is not None:
            if len(fsingle) != self.count_physical_qubit:
                raise ValueError( ("fsingle should have "
                                   f"{self.count_physical_qubit} data.") )
            self.list_fidelity_single = tuple(fsingle)
        if "list_fidelity_single" not in self.__dict__:
            self.list_fidelity_single = \
                tuple(1 for _ in range(self.count_physical_qubit))
        
        if ftwo is not None:
            if len(ftwo) != len(self.list_qubit_edge):
                raise ValueError( ("ftwo should have "
                                   f"{len(self.list_qubit_edge)} data.") )
            self.list_fidelity_two = tuple(ftwo)
        if "list_fidelity_two" not in self.__dict__:
            self.list_fidelity_two = \
                tuple(1 for _ in range(len(self.list_qubit_edge)))



from qiskit import QuantumCircuit

from olsq.solve import OLSQ
from olsq.device import qcdevice
from olsq.input import input_qasm


class OLSQ_qiskit(OLSQ):
    def __init__(self, objective_name, if_transition_based):
        """Set the objective of OLSQ_cirq, and whether transition-based.
        """

        super().__init__(objective_name, if_transition_based)

    def setdevice(self, device, mode: str = None):
        """Pass in a device representing hardware in Qiskit.

        Args:
            device: can be a qcdevice in OLSQ, or an IBM backend
            mode: if set to "ibm", process device as an IBM backend.
        """

        if mode == "ibm":
            config = device.configuration()
            edges = config.coupling_map # each edge has two tuples in this
            biedges = [] # represent each edge with one tuple

            for edge in edges:
                qubit0 = edge[0]
                qubit1 = edge[1]
                if qubit0 > qubit1:
                    tmp = qubit0
                    qubit0 = qubit1
                    qubit1 = tmp
                if (qubit0, qubit1) not in biedges:
                    biedges.append((qubit0, qubit1))

            super().setdevice(
                qcdevice(config.backend_name, nqubits=config.n_qubits,
                         connection=tuple(biedges), swap_duration=3    )  )
        else:
            super().setdevice(device)

    def setprogram(self, circuit: QuantumCircuit, input_mode: str = None):
        """Translate input Qiskit program/circuit to IR

        Args:
            circuit: a qasm string or a QuantumCircuit object in Qiskit
            input_mode: if set to "qasm", process curcyut as qasm
        """

        if input_mode == "qasm":
            program_qiskit = circuit
        else:
            circuit = circuit.decompose()
            program_qiskit = circuit.qasm()
        super().setprogram(program_qiskit)


    def solve(self):
        """Use the super().solve(...) and output in Qiskit

        Returns:
            circuit: result in Qiskit QuantumCircuit object
            final_mapping: from logical qubit to physical qubit
            objective_value: depth/#swap/fidelity depending on setting
        """

        circuit_str, final_mapping, objective_value = super().solve()
        circuit = QuantumCircuit()
        circuit = circuit.from_qasm_str(circuit_str)
        return [circuit, final_mapping, objective_value]


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="olsq",
    version="0.0.4.1",
    license = "BSD",
    author="Daniel Bochen Tan",
    author_email="bctan@cs.ucla.edu",
    description="Optimal Layout Synthesis for Quantum Computing (OLSQ) for mapping and scheduling quantum circuits",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tbcdebug/OLSQ",
    project_urls={
        "Bug Tracker": "https://github.com/tbcdebug/OLSQ/issues",
    },
    install_requires=[
        "networkx>=2.5",
        "z3-solver>=4.8.9.0",
    ],
    packages=setuptools.find_packages(),
    package_data={ "olsq": ["devices/*", "benchmarks/*"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)
